package Source::Repository;

use strict;
use warnings;

use File::Path qw(make_path remove_tree);
use File::Basename;
use Source::Package;

sub new {
    my $proto=shift;
    my $class=ref($proto) || $proto;
    my $self = {
	-cachedir=>'in',
	-path=>[],
	-names=>[],
	@_,
    };
    my $default_opt=$self->{-default_opt};
    if ($default_opt) {
	my @path=&Source::Repository::path_to_array($default_opt);
	if (not @{$self->{-path}} and not $self->{-pathlist} and grep {-d $_} @path) {
	    $self->{-path}=\@path;
	} elsif (not $self->{-repolist} and -f $default_opt) {
	    $self->{-repolist}=$default_opt;
	}
    }
    if ($self->{-pathlist}) {
	push @{$self->{-path}}, &Source::Repository::path_to_array($self->{-pathlist});
    }
    bless($self,$class);
    $self->__load_repolist();
    $self->__load_namelist();
    $self->init_hook();
    die "internal error: invalid repository path" if not ref $self->{-path};
    die "internal error: invalid list of names" if not ref $self->{-names};

    die "can't initialize repository: neither path nor list is specified" if not $self->{-name2pkg} and not scalar @{$self->{-path}};
    return $self;
}

sub init_hook {}

sub expand_glob {
    my ($self,$item)=@_;
    return &__expand_glob($item,$self->{-names});
}

sub names {
    my ($self)=@_;
    return @{$self->{-names}};
}

sub download {
    my ($self,$pkg)=@_;
    my $file=$pkg->file;
    return $file if defined $file and -e $file and -s $file;
    my $url=$pkg->url;
    my $filename=basename($url);
    $filename=~s/\?.*$//;
    my $cachedir=$self->{-cachedir};
    make_path($cachedir);
    $file="$cachedir/$filename";
    return $file if -e $file and -s $file;
    unlink $file;
    `wget -nv -O $file $url`;
    unlink $file if -z $file;
    die "download failed" unless -e $file;
    $pkg->{-file}=$file;
    return $file;
}

sub pkg {
    my ($self,$item)=@_;
    return $self->file2pkg($item) if defined $item and -e $item;
    return $self->name2pkg($item);
}

sub name2pkg {
    my ($self,$name)=@_;
    my $name2pkg=$self->{-name2pkg};
    if ($name2pkg) {
	return $name2pkg->{$name};
    } elsif ($self->{-findsub}) {
	my $file=&{$self->{-findsub}}($name,@{$self->{-path}});
	return unless $file;
	return $self->file2pkg();
    }
    return;
}

sub __load_repolist {
    my $self=shift;
    my $list=$self->{-repolist};
    return unless defined $list;
    ($self->{-names}, $self->{-name2pkg})=&__load_list($list,\&default_list3_processor,$self);

}

sub __load_namelist {
    my $self=shift;
    my $list=$self->{-namelist};
    return unless defined $list;
    ($self->{-names})=&__load_list($list);
}

#------------------------------------------
#    protected utilities
#------------------------------------------

# used in Source::Repository::Mass
sub __line2array {
    my $line=shift;
    chomp $line;
    $line=~s/^\s*//;
    $line=~s/\s*$//;
    return if $line=~/^\s*#/;
    return if $line eq '';
    return split (/\s+/, $line);
}

# used in RPM
sub path_to_array {
    my $paths=shift;
    my @path;
    foreach (split(':',$paths)) {
	if ($_ eq '') {
	    $_ = '.';
	} elsif (/^~/) {
	    s/^~/$ENV{HOME}/;
	}
	push @path, $_ if -d $_;
    }
    return @path;
}

#------------------------------------------
#    private utilities
#------------------------------------------

# glob expansion on $opt_namelist
sub __expand_glob {
    my ($pattern, $arrayptr)=@_;
    return $pattern unless $pattern =~/[?*]/;
    my $regexp=&glob2pat($pattern);
    my @names = grep {/$regexp/} @$arrayptr;
    print STDERR "WARNING: pattern $pattern does not match.\n" unless @names;
    return @names;
}

sub glob2pat {
    my $globstr = shift;
    my %patmap = (
	'*' => '.*',
	'?' => '.',
	'[' => '[',
	']' => ']',
	'{' => '(?:',
	'}' => ')',
	',' => '|',
	);
    $globstr =~ s{(.)} { $patmap{$1} || "\Q$1" }ge;
    return '^' . $globstr . '$';
}

sub __load_list {
    my $list=shift;
    my $processor=shift;
    return unless defined $list;
    my @all_names;
    my %name2pkg;
    open (my $fn, '<', $list) or die "can't open file $list\n";
    while (my $line=<$fn>) {
	my @list = &__line2array($line);
	next unless @list;
	my $name=$list[0];
	push @all_names, $name;
	$name2pkg{$name}=&$processor(\@list,@_) if $processor;
    }
    close($fn);
    return \@all_names, \%name2pkg;
}

sub default_list3_processor {
    my ($ref,$self)=@_;
    return unless $ref;
    my ($epoch,$release);
    return unless @$ref;
    my ($name, $versionspec, $url)=@$ref;
    $epoch=$1 if $versionspec=~s/^(\d+)://;
    # warning: altlinux-specific
    $release=$1 if $versionspec=~s/-(alt.+)$//;
    # TODO: baseurl support
    return Source::Package->new(
	-name => $name,
	-epoch=> $epoch,
	-version => $versionspec,
	-release => $release,
	-url => $url,
	);
}
1;
