#!/usr/bin/perl -w

package Source::Repository::Mass;

use strict;
use warnings;
use Cwd;
use Carp;
use Getopt::Long;
use File::Basename;
use File::Path qw(make_path remove_tree);
use RPM qw/evrcmp vercmp/;

use Source::Repository::RPM;
use Source::Repository::Matcher;

require Exporter;

our $VERSION=0.04;

our @ISA = qw(Exporter);
our @EXPORT = qw/$DESTPATH $ORIGINPATH @ORIGINSKIP @DESTSKIP/;
our ($DESTPATH, $ORIGINPATH);

our @ORIGINSKIP=();
our @DESTSKIP=();
my %ORIGINSKIP;
my %DESTSKIP;

our $matcher_class='Source::Repository::Matcher';
our $origin_class='Source::Repository::RPM';
our $dest_class='Source::Repository::RPM';

our $OUTDIR="./OUT";
our $verbose=0;
our $strategy = 'all';

my ($opt_transaction, $opt_taskfile, $opt_all, $opt_destmode, $help);
my ($opt_origin_namelist, $opt_dest_namelist, $opt_namelist);
my ($opt_origin_repolist, $opt_dest_repolist);
my ($opt_origin_pathlist, $opt_dest_pathlist);
my ($opt_origin_cachedir, $opt_dest_cachedir);
# TODO
my ($opt_origin_skiplist, $opt_dest_skiplist);

our ($arg_timestampdir,$opt_skip_on_error);
our $arg_timestamp_expire=7; # in days


our %STRATEGYMAP=(
    'all' => \&origin2dest,
    'missing' => \&originskipdest,
    'orphaned' => \&destskiporigin,
    'destup' => \&destup,
    'newer' => \&newer,
    'older' => \&older,
    # old names
    'origin2dest' => \&origin2dest,
    'originskipdest' => \&originskipdest,
    'destskiporigin' => \&destskiporigin,
    );

our %STRATEGYDESCR=(
    'all' => 'any package',
    'missing' => 'packages in origin, missing in dest', 
    'orphaned' => 'packages in dest, missing in origin',
    'destup' => 'packages in dest, older than origin (i.e. need update)',
    'newer' => 'packages in origin, new and newer than dest',
    'older' => 'packages in origin, older than dest',
    );

my @LONGOPT = (
	"all"	=> \$opt_all,
	"dest=s"	=> \$DESTPATH,
	"dest-cachedir=s"	=> \$opt_dest_cachedir,
	"dest-namelist=s"	=> \$opt_dest_namelist,
	"dest-repolist=s"	=> \$opt_dest_repolist,
	"dest-path=s"		=> \$opt_dest_pathlist,
	"dest-skiplist=s"	=> \$opt_dest_skiplist,
	"input-dest"	=> sub {$opt_destmode=1},
	"input-origin"	=> sub {$opt_destmode=0},
	"origin=s"	=> \$ORIGINPATH,
	"origin-cachedir=s"	=> \$opt_origin_cachedir,
	"origin-namelist=s"	=> \$opt_origin_namelist,
	"origin-repolist=s"	=> \$opt_origin_repolist,
	"origin-path=s"		=> \$opt_origin_pathlist,
	"origin-skiplist=s"	=> \$opt_origin_skiplist,
	"out|outdir=s"	=> \$OUTDIR,
	"skip-on-error"	=> \$opt_skip_on_error,
	"strategy=s"	=> \$strategy,
	"timestampdir=s"	=> \$arg_timestampdir,
	"timestamp-expire=f"	=> \$arg_timestamp_expire,
	"transaction"	=> \$opt_transaction,
	"tasklist=s"	=> \$opt_taskfile,
	"namelist=s"	=> \$opt_namelist,

	"help"		=> \$help,
	"v|verbose+"	=> \$verbose,
	'q|quiet'   => sub { $verbose = 0 },
    );
our @EXT_OPTS;
our @EXT_OPTS_USAGE;

sub new {
    my $class = shift;
    my $self = {
	'-originopts' => [],
	'-destopts' => [],
	'-cliargs' => [],
	'-packages' => [],
	@_
    };
    bless $self, $class;
    my $result = GetOptions (
	@LONGOPT,
	@EXT_OPTS,
	@{$self->{'-cliargs'}},
	);

    if (not $result) {
	print "run $0 --help for details.\n";
	exit 1;
    } elsif ($help) {
	$self->usage();
	exit 0;
    }

    $self->{VERBOSE}=$verbose;

    my (@DEFREPO_OPTS,@ORIGIN_OPTS,@DEST_OPTS);
    push @DEFREPO_OPTS, -namelist => $opt_namelist if $opt_namelist;
    push @ORIGIN_OPTS, -namelist => $opt_origin_namelist if $opt_origin_namelist;
    push @DEST_OPTS, -namelist => $opt_dest_namelist if $opt_dest_namelist;
    push @ORIGIN_OPTS, -repolist => $opt_origin_repolist if $opt_origin_repolist;
    push @DEST_OPTS, -repolist => $opt_dest_repolist if $opt_dest_repolist;
    push @ORIGIN_OPTS, -cachedir => $opt_origin_cachedir if $opt_origin_cachedir;
    push @DEST_OPTS, -cachedir => $opt_dest_cachedir if $opt_dest_cachedir;
    push @ORIGIN_OPTS, -pathlist => $opt_origin_pathlist if $opt_origin_pathlist;
    push @DEST_OPTS, -pathlist => $opt_dest_pathlist if $opt_dest_pathlist;

    $self->{-originrepo}=$origin_class->new(
	-default_opt=>$ORIGINPATH,
 	-rpmfindsub => \&Source::Repository::RPM::find_first_rpm,
	$opt_destmode ? () : @DEFREPO_OPTS,
	@ORIGIN_OPTS,
	@{$self->{-originopts}},
	);
    $self->{-destrepo}=$dest_class->new(
	-default_opt=>$DESTPATH,
	-rpmfindsub => \&Source::Repository::RPM::find_highest_rpm,
	$opt_destmode ? @DEFREPO_OPTS : (),
	@DEST_OPTS,
	@{$self->{-destopts}},
	);

    $self->{-matcher} = $matcher_class->new(
	-originrepo => $self->{-originrepo}, 
	-destrepo => $self->{-destrepo}, 
	-destmode=>$opt_destmode,
	);

    %ORIGINSKIP=map {$_ => 1} @ORIGINSKIP;
    %DESTSKIP=map {$_ => 1} @DESTSKIP;

    my $def_repo = $opt_destmode ? $self->{-destrepo} : $self->{-originrepo};
    if ($opt_taskfile) {
	open (my $fn, '<', $opt_taskfile) or die "can't open file $opt_taskfile\n";
	while (my $line=<$fn>) {
	    my @list = map{$def_repo->expand_glob($_)} &Source::Repository::__line2array($line);
	    next if scalar @list == 0;
	    if (scalar @list == 1) {
		push @{$self->{'-packages'}}, @list;
	    } else {
		# transaction
		push @{$self->{'-packages'}}, [@list];
	    }
	}
	close($fn);
    }

    make_path($arg_timestampdir) if defined $arg_timestampdir;
    #make_path ($OUTDIR) unless -d $OUTDIR;

    $self->process_options_hook();

    push @{$self->{'-packages'}}, $opt_transaction ? [@ARGV] : @ARGV;
    push @{$self->{'-packages'}}, $def_repo->names() if $opt_all;
    return $self;
}

sub process_options_hook {
    my ($self)=@_;
}

sub usage {
    print "usage: $0 [ options ] [ list of origin rpms/names ]

Example:
upgrade all dest java packages using jpackage
originmass --dest /var/ftp/pub/Sisyphus/files/SRPMS --origin /var/ftp/pub/jpackage/1.7/generic/free/SRPMS

General options:
	-h, --help
	-v, --verbose (can be specified multiple times)
	-q, --quiet

Mass processing options:
	--origin	/path1/to/foreign|origin/src/RPMS:/other/path2
	--dest	/path1/to/altlinux|dest/src/RPMS:/other/path1:/other/path2
	--out	/path/to/output_dir (default is $OUTDIR)
	--skip-errors	ignore individual package failures
	--strategy 'strategy'	of mass operations (default is $strategy).
	  Available strategies: 
";
    foreach my $sname (sort keys(%STRATEGYMAP)) {
	my $description=$STRATEGYDESCR{$sname};
	print "\t\t$sname\t- $description\n" if defined $description;
    }

    foreach my $usage(@EXT_OPTS_USAGE) {
	if (ref $usage ne 'CODE') {
	    die "Oops! not a code: $usage ".(ref $usage eq 'SCALAR' ? $$usage : '');
	} else {
	    print "\n";
	    &$usage;
	}
    }
}

sub originrepo {
    my ($self)=@_;
    return $self->{-originrepo};
}

sub destrepo {
    my ($self)=@_;
    return $self->{-destrepo};
}

sub process {
    my ($self)=@_;
    foreach my $nameobj (@{$self->{'-packages'}}) {
	if (ref $nameobj) {
	    $self->_process_transaction($nameobj);
	} else {
	    $self->_process_name($nameobj);
	}
    }
}

sub _process_name {
    my ($self, $name)=@_;
    my $pkgpair=$self->{-matcher}->new_pair($name);
    return if $self->__skip_due_to_timestamp($pkgpair);
    return unless $self->is_strategy_allow($pkgpair);
    $self->__mk_timestamp($pkgpair);
    $self->_process_pair_wrapper($pkgpair);
}

sub _process_transaction {
    my ($self, $transaction_ptr)=@_;
    my $transaction_name=$transaction_ptr->[0];
    my $is_in_transaction=0;
    my $transaction_count=0;
    foreach my $name (@$transaction_ptr) {
	my $pkgpair=$self->{-matcher}->new_pair($name);
	next if $self->__skip_due_to_timestamp($pkgpair);
	next unless $self->is_strategy_allow($pkgpair);
	if (not $is_in_transaction) {
	    $self->init_transaction($transaction_name);
	    $is_in_transaction=1;
	}
	$self->__mk_timestamp($pkgpair);
	$self->_process_pair_wrapper($pkgpair, $transaction_name, ++$transaction_count);
    }
    $self->done_transaction($transaction_name,$transaction_count) if $is_in_transaction;
}

sub init_transaction {
    my ($self,$transaction_name)=@_;
}

sub done_transaction {
    my ($self,$transaction_name)=@_;
}

sub _process_pair_wrapper {
    my ($self,$pkgpair,$transaction_name,$transaction_count)=@_;
    eval {
	$self->process_pair($pkgpair,$transaction_name,$transaction_count);
    };
    if ($@) {
	die 'FAILED: '.$pkgpair->originname.": $@

You can ignore this error with --skip-on-error option\n" unless $opt_skip_on_error;
	print STDERR 'FAILED: '.$pkgpair->originname.": $@";
    };
}

###############################################################
# The method process_pair should be implemented in subclasses;
# disable the prototype implementation completely to allow 
# multiple inheritance tricks with Transform::{Embedded,PassThrough}
###############################################################
#sub process_pair {
#    my ($self,$pkgpair,$transaction_name,$transaction_count)=@_;
#    print STDERR "process_pair: $self\n";
#    ...
#}

sub is_strategy_allow {
    my ($self,$pkgpair)=@_;
    my $strategy_allow=$STRATEGYMAP{$strategy};
    die "strategy $strategy not found" unless $strategy_allow;
    my $originname=$pkgpair->originname;
    my $destname=$pkgpair->destname;
    return 0 if $ORIGINSKIP{$originname} or $DESTSKIP{$destname};
    return 1 if &{$strategy_allow}($self, $pkgpair->origin, $pkgpair->dest, $originname, $destname);
    return 0;
}

sub cmp_origin_dest {
    my ($self, $origin, $dest)=@_;
    return 0 unless $origin;
    return 1 unless $dest;
    return vercmp($origin->version, $origin->release, $dest->version, $dest->release)
}

sub __skip_due_to_timestamp {
    my ($self,$pkgpair)=@_;
    return 0 unless defined $arg_timestampdir;
    my $timestamp_file=&__srpmfile2timestamp($pkgpair->origin->file);
    return 1 if -M $timestamp_file and ($arg_timestamp_expire==0 or -M $timestamp_file < $arg_timestamp_expire);
    return 0;
}

sub __mk_timestamp {
    my ($self,$pkgpair)=@_;
    return unless defined $arg_timestampdir;
    &__run('touch',&__srpmfile2timestamp($pkgpair->origin->file));
}

#------------------------------------------

#------------------------------------------
#    private strategies
#------------------------------------------

sub origin2dest {
    my ($self,$originname, $destname,$origin, $dest)=@_;
    print STDERR "FATAL: origin RPM $originname not found.\n" unless $origin;
    return 0 unless $origin;
    print STDERR "note: dest RPM $destname not found.\n" unless $dest;
    return 1;
}

sub destup {
    my ($self, $origin, $dest, $originname, $destname)=@_;
    print STDERR "FATAL: origin RPM $originname not found.\n" unless $origin;
    return 0 unless $origin;
    print STDERR "note: dest RPM $destname not found.\n" unless $dest;
    return 1 if $dest and $self->cmp_origin_dest($origin,$dest)>0;
    return 0;
}

sub newer {
    my ($self, $origin, $dest, $originname, $destname)=@_;
    print STDERR "FATAL: origin RPM $originname not found.\n" unless $origin;
    return 0 unless $origin;
    print STDERR "note: origin RPM $originname not found.\n" unless $origin;
    return 1 if not $dest or $self->cmp_origin_dest($origin,$dest)>0;
    return 0;
}

sub originskipdest {
    my ($self, $origin, $dest, $originname, $destname)=@_;
    print STDERR "FATAL: origin RPM $originname not found.\n" unless $origin;
    return 0 unless $origin;
    print STDERR "found dest RPM $destname for $originname.\n" if $dest;
    return 0 if $dest;
    return 1;
}

sub older {
    my ($self, $origin, $dest, $originname, $destname)=@_;
    print STDERR "FATAL: origin RPM $originname not found.\n" unless $origin;
    return 0 unless $origin;
    #print STDERR "note: dest RPM $destname not found.\n" unless $dest;
    return 1 if $dest and $self->cmp_origin_dest($origin,$dest)<0;
    return 0;
}

sub destskiporigin {
    my ($self, $origin, $dest, $originname, $destname)=@_;
    print STDERR "FATAL: dest RPM $destname not found.\n" unless $dest;
    return 0 unless $dest;
    print STDERR "found origin RPM $originname for $destname.\n" if $origin;
    return 0 if $origin;
    return 1;
}


#------------------------------------------

#------------------------------------------
#    private utilities
#------------------------------------------

sub __srpmfile2timestamp {
    my ($rpm)=@_;
    return $arg_timestampdir.'/'.basename($rpm).'.stamp';
}

sub __run {
    print STDERR "running: ", join(' ',@_),"\n" if $Source::Repository::Mass::verbose;
    system(@_)==0 or Carp::confess "command failed: ".join(' ',@_);
}


1;
